#include "anchor_constraints.hh"
#include <assert.h>
#include <sstream>

#include <iostream>

AnchorConstraints::failure::~failure() throw() {};

const char *
AnchorConstraints::failure::what() const throw() {
    //exception::what();
    return msg_.c_str();
}

AnchorConstraints::AnchorConstraints(size_type lenA,
			       const std::vector<std::string> &seqVecA,
			       size_type lenB,
			       const std::vector<std::string> &seqVecB)
    : a(lenA+1,0),
      b(lenB+1,0),
      ar(lenA+1,range_t(1,lenB))
{
    name_size_=seqVecA.size();

    if (seqVecA.size()!=seqVecB.size()) {
	throw( failure("Wrong input for sequence constraints. Lengths of names in sequences don't fit.") );
    }

    std::map<std::string,size_type> nameTabA;
    std::map<std::string,size_type> nameTabB;
   
    transform_input(nameTabA,lenA,seqVecA);
    transform_input(nameTabB,lenB,seqVecB);

    init_tables(nameTabA,nameTabB);
}


AnchorConstraints::AnchorConstraints(size_type lenA,
			       const std::string &seqCA,
			       size_type lenB,
			       const std::string &seqCB)
    : a(lenA+1,0),
      b(lenB+1,0),
      ar(lenA+1,range_t(1,lenB)) 
{
    std::vector<std::string> seqVecA;
    std::vector<std::string> seqVecB;
    
    transform_input(seqVecA,seqCA);
    transform_input(seqVecB,seqCB);

    if (seqVecA.size()!=seqVecB.size()) {
	throw( failure("Error during parsing of constraints. Lengths of names in sequences don't fit.") );
    }

    name_size_=seqVecA.size();
    
    std::map<std::string,size_type> nameTabA;
    std::map<std::string,size_type> nameTabB;
    
    transform_input(nameTabA,lenA,seqVecA);
    transform_input(nameTabB,lenB,seqVecB);
    
    init_tables(nameTabA,nameTabB);
}

void AnchorConstraints::transform_input(std::vector<std::string> &seqVec,
				     const std::string &seqStr) {
    std::string line;
    std::istringstream ss(seqStr);
    while (getline(ss,line,'#')) {
	seqVec.push_back(line);
    }
}

void
AnchorConstraints::transform_input(name_tab_t &nameTab,
				size_type len,
				const std::vector<std::string> &seq) {
    std::vector<std::string> vec(len,"");
    
    for(std::vector<std::string>::const_iterator it=seq.begin();
	seq.end() != it;
	++it) {
	if (len != it->length()) 
	    throw( failure("Error during parsing of constraints. Constraint string of wrong length.") );
	
	for (std::string::size_type i=0; i<len; i++) {
	    vec[i].push_back((*it)[i]);
	}
    }

    size_type i=1;
    for(std::vector<std::string>::iterator it=vec.begin();
	vec.end()!=it;
	++it) {
	nameTab[*it]=i;
	++i;
    }
}

void
AnchorConstraints::init_seq_table(seq_t & seq_tab,
			       name_seq_t & name_seq_tab,
			       const name_tab_t &nameTabA,
			       const name_tab_t &nameTabB) {
    for (name_tab_t::const_iterator it=nameTabA.begin();
	 nameTabA.end()!=it;
	 ++it) {
	std::string name=it->first;
	size_type posA=it->second;
	
	name_seq_tab[posA] = name;
	
	name_tab_t::const_iterator itB = nameTabB.find(name);
	
	if (itB != nameTabB.end()) {
	    size_type posB = itB->second;
	    seq_tab[posA] = posB;
	} else {
	    seq_tab[posA] = -1;
	}
    }
}

void
AnchorConstraints::init_tables(const name_tab_t &nameTabA,
			    const name_tab_t &nameTabB) {
    
    size_type lenA = a.size()-1; // -1 !!! 
    
    names_a.resize(a.size());
    names_b.resize(b.size());
    
    // named positions a
    init_seq_table(a,names_a,nameTabA,nameTabB);
    
    // named positions b
    init_seq_table(b,names_b,nameTabB,nameTabA); // (symmetrical call)
    
    size_type last=0;
    
    // matches from a to b
    
    /*
      scan A twice.
      First, from left to right and set left end of range.
      Second, from right to left and set right end.
     */
    
    for (size_type i=0; i<=lenA; i++) {
	if (a[i] > 0) {
	    last  = a[i];
	    ar[i].first = last;
	} else {
	    ar[i].first = last+1;
	}
    }
    
    last = b.size()+1;
    for (size_type i=lenA; i>=1; i--) {
	if (a[i] > 0) {
	    last  = a[i];
	    ar[i].second = last;
	} else {
	    ar[i].second = last-1;
	}
    }
}


